@@ -3,7 +3,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2017 David R. MacIver
+# Most of this work is copyright (C) 2013-2018 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -23,17 +23,25 @@ import numpy as np
 
 import hypothesis.strategies as st
 import hypothesis.internal.conjecture.utils as cu
+from hypothesis import Verbosity
 from hypothesis.errors import InvalidArgument
+from hypothesis._settings import note_deprecation
+from hypothesis.reporting import current_verbosity
 from hypothesis.searchstrategy import SearchStrategy
 from hypothesis.internal.compat import hrange, text_type
 from hypothesis.internal.coverage import check_function
 from hypothesis.internal.reflection import proxies
 
+if False:
+    from typing import Any, Union, Sequence, Tuple  # noqa
+    from hypothesis.searchstrategy.strategies import T  # noqa
+
 TIME_RESOLUTIONS = tuple('Y  M  D  h  m  s  ms  us  ns  ps  fs  as'.split())
 
 
 @st.defines_strategy_with_reusable_values
 def from_dtype(dtype):
+    # type: (np.dtype) -> st.SearchStrategy[Any]
     # Compound datatypes, eg 'f4,f4,f4'
     if dtype.names is not None:
         # mapping np.void.type over a strategy is nonsense, so return now.
@@ -47,21 +55,35 @@ def from_dtype(dtype):
 
     # Scalar datatypes
     if dtype.kind == u'b':
-        result = st.booleans()
+        result = st.booleans()  # type: SearchStrategy[Any]
     elif dtype.kind == u'f':
-        result = st.floats()
+        if dtype.itemsize == 2:
+            result = st.floats(width=16)
+        elif dtype.itemsize == 4:
+            result = st.floats(width=32)
+        else:
+            result = st.floats()
     elif dtype.kind == u'c':
-        result = st.complex_numbers()
+        if dtype.itemsize == 8:
+            float32 = st.floats(width=32)
+            result = st.builds(complex, float32, float32)
+        else:
+            result = st.complex_numbers()
     elif dtype.kind in (u'S', u'a'):
-        result = st.binary()
+        # Numpy strings are null-terminated; only allow round-trippable values.
+        # `itemsize == 0` means 'fixed length determined at array creation'
+        result = st.binary(max_size=dtype.itemsize or None
+                           ).filter(lambda b: b[-1:] != b'\0')
     elif dtype.kind == u'u':
-        result = st.integers(
-            min_value=0, max_value=1 << (4 * dtype.itemsize) - 1)
+        result = st.integers(min_value=0,
+                             max_value=2 ** (8 * dtype.itemsize) - 1)
     elif dtype.kind == u'i':
-        min_integer = -1 << (4 * dtype.itemsize - 1)
-        result = st.integers(min_value=min_integer, max_value=-min_integer - 1)
+        overflow = 2 ** (8 * dtype.itemsize - 1)
+        result = st.integers(min_value=-overflow, max_value=overflow - 1)
     elif dtype.kind == u'U':
-        result = st.text()
+        # Encoded in UTF-32 (four bytes/codepoint) and null-terminated
+        result = st.text(max_size=(dtype.itemsize or 0) // 4 or None
+                         ).filter(lambda b: b[-1:] != u'\0')
     elif dtype.kind in (u'm', u'M'):
         if '[' in dtype.str:
             res = st.just(dtype.str.split('[')[-1][:-1])
@@ -86,7 +108,7 @@ def order_check(name, floor, small, large):
         floor, small, name=name
     )
     check_argument(
-        small <= large, u'min_{name}={} is larger than max_name={}',
+        small <= large, u'min_{name}={} is larger than max_{name}={}',
         small, large, name=name
     )
 
@@ -107,10 +129,44 @@ class ArrayStrategy(SearchStrategy):
         self.element_strategy = element_strategy
         self.unique = unique
 
+        # Used by self.insert_element to check that the value can be stored
+        # in the array without e.g. overflowing.  See issue #1385.
+        if dtype.kind in (u'i', u'u'):
+            self.check_cast = lambda x: np.can_cast(x, self.dtype, 'safe')
+        elif dtype.kind == u'f' and dtype.itemsize == 2:
+            max_f2 = (2. - 2 ** -10) * 2 ** 15
+            self.check_cast = lambda x: \
+                (not np.isfinite(x)) or (-max_f2 <= x <= max_f2)
+        elif dtype.kind == u'f' and dtype.itemsize == 4:
+            max_f4 = (2. - 2 ** -23) * 2 ** 127
+            self.check_cast = lambda x: \
+                (not np.isfinite(x)) or (-max_f4 <= x <= max_f4)
+        else:
+            self.check_cast = lambda x: True
+
+    def set_element(self, data, result, idx, strategy=None):
+        strategy = strategy or self.element_strategy
+        val = data.draw(strategy)
+        if self._report_overflow and not self.check_cast(val):
+            note_deprecation(
+                'Generated array element %r from %r cannot be represented as '
+                'dtype %r without overflow or underflow.  Consider using a '
+                'more precise strategy, as this will be an error in a future '
+                'version of Hypothesis.' % (val, strategy, self.dtype)
+            )
+            # Because the message includes the value of the generated element,
+            # it would be easy to spam users with thousands of warnings.
+            # We therefore only warn once per draw, unless in verbose mode.
+            self._report_overflow = current_verbosity() >= Verbosity.verbose
+        result[idx] = val
+
     def do_draw(self, data):
         if 0 in self.shape:
             return np.zeros(dtype=self.dtype, shape=self.shape)
 
+        # Reset this flag for each test case to emit warnings from set_element
+        self._report_overflow = True
+
         # This could legitimately be a np.empty, but the performance gains for
         # that would be so marginal that there's really not much point risking
         # undefined behaviour shenanigans.
@@ -135,7 +191,7 @@ class ArrayStrategy(SearchStrategy):
                     # uniqueness after numpy has converted it to the relevant
                     # type for us. Because we don't increment the counter on
                     # a duplicate we will overwrite it on the next draw.
-                    result[i] = data.draw(self.element_strategy)
+                    self.set_element(data, result, i)
                     if result[i] not in seen:
                         seen.add(result[i])
                         i += 1
@@ -143,7 +199,7 @@ class ArrayStrategy(SearchStrategy):
                         elements.reject()
             else:
                 for i in hrange(len(result)):
-                    result[i] = data.draw(self.element_strategy)
+                    self.set_element(data, result, i)
         else:
             # We draw numpy arrays as "sparse with an offset". We draw a
             # collection of index assignments within the array and assign
@@ -169,7 +225,7 @@ class ArrayStrategy(SearchStrategy):
                 if not needs_fill[i]:
                     elements.reject()
                     continue
-                result[i] = data.draw(self.element_strategy)
+                self.set_element(data, result, i)
                 if self.unique:
                     if result[i] in seen:
                         elements.reject()
@@ -190,7 +246,7 @@ class ArrayStrategy(SearchStrategy):
                 # it and putmask will do the right thing by repeating the
                 # values of the array across the mask.
                 one_element = np.zeros(shape=1, dtype=self.dtype)
-                one_element[0] = data.draw(self.fill)
+                self.set_element(data, one_element, 0, self.fill)
                 fill_value = one_element[0]
                 if self.unique:
                     try:
@@ -222,8 +278,14 @@ def fill_for(elements, unique, fill, name=''):
 
 @st.composite
 def arrays(
-    draw, dtype, shape, elements=None, fill=None, unique=False
+    draw,  # type: Any
+    dtype,  # type: Any
+    shape,  # type: Union[int, Sequence[int]]
+    elements=None,  # type: st.SearchStrategy[Any]
+    fill=None,  # type: st.SearchStrategy[Any]
+    unique=False,  # type: bool
 ):
+    # type: (...) -> st.SearchStrategy[np.ndarray]
     """Returns a strategy for generating :class:`numpy's
     ndarrays<numpy.ndarray>`.
 
@@ -293,7 +355,6 @@ def arrays(
     primarily designed around testing small examples. If you have arrays with
     hundreds or more elements, having a fill value is essential if you want
     your tests to run in reasonable time.
-
     """
     if isinstance(dtype, SearchStrategy):
         dtype = draw(dtype)
@@ -316,6 +377,7 @@ def arrays(
 
 @st.defines_strategy
 def array_shapes(min_dims=1, max_dims=3, min_side=1, max_side=10):
+    # type: (int, int, int, int) -> st.SearchStrategy[Tuple[int, ...]]
     """Return a strategy for array shapes (tuples of int >= 1)."""
     order_check('dims', 1, min_dims, max_dims)
     order_check('side', 1, min_side, max_side)
@@ -325,6 +387,7 @@ def array_shapes(min_dims=1, max_dims=3, min_side=1, max_side=10):
 
 @st.defines_strategy
 def scalar_dtypes():
+    # type: () -> st.SearchStrategy[np.dtype]
     """Return a strategy that can return any non-flexible scalar dtype."""
     return st.one_of(boolean_dtypes(),
                      integer_dtypes(), unsigned_integer_dtypes(),
@@ -333,6 +396,7 @@ def scalar_dtypes():
 
 
 def defines_dtype_strategy(strat):
+    # type: (T) -> T
     @st.defines_strategy
     @proxies(strat)
     def inner(*args, **kwargs):
@@ -342,6 +406,7 @@ def defines_dtype_strategy(strat):
 
 @defines_dtype_strategy
 def boolean_dtypes():
+    # type: () -> st.SearchStrategy[np.dtype]
     return st.just('?')
 
 
@@ -370,6 +435,7 @@ def dtype_factory(kind, sizes, valid_sizes, endianness):
 
 @defines_dtype_strategy
 def unsigned_integer_dtypes(endianness='?', sizes=(8, 16, 32, 64)):
+    # type: (str, Sequence[int]) -> st.SearchStrategy[np.dtype]
     """Return a strategy for unsigned integer dtypes.
 
     endianness may be ``<`` for little-endian, ``>`` for big-endian,
@@ -378,24 +444,24 @@ def unsigned_integer_dtypes(endianness='?', sizes=(8, 16, 32, 64)):
 
     sizes must be a collection of integer sizes in bits.  The default
     (8, 16, 32, 64) covers the full range of sizes.
-
     """
     return dtype_factory('u', sizes, (8, 16, 32, 64), endianness)
 
 
 @defines_dtype_strategy
 def integer_dtypes(endianness='?', sizes=(8, 16, 32, 64)):
+    # type: (str, Sequence[int]) -> st.SearchStrategy[np.dtype]
     """Return a strategy for signed integer dtypes.
 
     endianness and sizes are treated as for
     :func:`unsigned_integer_dtypes`.
-
     """
     return dtype_factory('i', sizes, (8, 16, 32, 64), endianness)
 
 
 @defines_dtype_strategy
 def floating_dtypes(endianness='?', sizes=(16, 32, 64)):
+    # type: (str, Sequence[int]) -> st.SearchStrategy[np.dtype]
     """Return a strategy for floating-point dtypes.
 
     sizes is the size in bits of floating-point number.  Some machines support
@@ -404,19 +470,18 @@ def floating_dtypes(endianness='?', sizes=(16, 32, 64)):
     Larger floats (96 and 128 bit real parts) are not supported on all
     platforms and therefore disabled by default.  To generate these dtypes,
     include these values in the sizes argument.
-
     """
     return dtype_factory('f', sizes, (16, 32, 64, 96, 128), endianness)
 
 
 @defines_dtype_strategy
 def complex_number_dtypes(endianness='?', sizes=(64, 128)):
+    # type: (str, Sequence[int]) -> st.SearchStrategy[np.dtype]
     """Return a strategy for complex-number dtypes.
 
     sizes is the total size in bits of a complex number, which consists
     of two floats.  Complex halfs (a 16-bit real part) are not supported
     by numpy and will not be generated by this strategy.
-
     """
     return dtype_factory('c', sizes, (64, 128, 192, 256), endianness)
 
@@ -439,6 +504,7 @@ def validate_time_slice(max_period, min_period):
 
 @defines_dtype_strategy
 def datetime64_dtypes(max_period='Y', min_period='ns', endianness='?'):
+    # type: (str, str, str) -> st.SearchStrategy[np.dtype]
     """Return a strategy for datetime64 dtypes, with various precisions from
     year to attosecond."""
     return dtype_factory('datetime64[{}]',
@@ -448,6 +514,7 @@ def datetime64_dtypes(max_period='Y', min_period='ns', endianness='?'):
 
 @defines_dtype_strategy
 def timedelta64_dtypes(max_period='Y', min_period='ns', endianness='?'):
+    # type: (str, str, str) -> st.SearchStrategy[np.dtype]
     """Return a strategy for timedelta64 dtypes, with various precisions from
     year to attosecond."""
     return dtype_factory('timedelta64[{}]',
@@ -457,6 +524,7 @@ def timedelta64_dtypes(max_period='Y', min_period='ns', endianness='?'):
 
 @defines_dtype_strategy
 def byte_string_dtypes(endianness='?', min_len=0, max_len=16):
+    # type: (str, int, int) -> st.SearchStrategy[np.dtype]
     """Return a strategy for generating bytestring dtypes, of various lengths
     and byteorder."""
     order_check('len', 0, min_len, max_len)
@@ -466,6 +534,7 @@ def byte_string_dtypes(endianness='?', min_len=0, max_len=16):
 
 @defines_dtype_strategy
 def unicode_string_dtypes(endianness='?', min_len=0, max_len=16):
+    # type: (str, int, int) -> st.SearchStrategy[np.dtype]
     """Return a strategy for generating unicode string dtypes, of various
     lengths and byteorder."""
     order_check('len', 0, min_len, max_len)
@@ -474,23 +543,34 @@ def unicode_string_dtypes(endianness='?', min_len=0, max_len=16):
 
 
 @defines_dtype_strategy
-def array_dtypes(subtype_strategy=scalar_dtypes(),
-                 min_size=1, max_size=5, allow_subarrays=False):
+def array_dtypes(
+    subtype_strategy=scalar_dtypes(),  # type: st.SearchStrategy[np.dtype]
+    min_size=1,  # type: int
+    max_size=5,  # type: int
+    allow_subarrays=False,  # type: bool
+):
+    # type: (...) -> st.SearchStrategy[np.dtype]
     """Return a strategy for generating array (compound) dtypes, with members
     drawn from the given subtype strategy."""
     order_check('size', 0, min_size, max_size)
-    native_strings = st.text if text_type is str else st.binary
-    elements = st.tuples(native_strings(), subtype_strategy)
+    native_strings = st.text()  # type: SearchStrategy[Any]
+    if text_type is not str:  # pragma: no cover
+        native_strings = st.binary()
+    elements = st.tuples(native_strings, subtype_strategy)
     if allow_subarrays:
-        elements |= st.tuples(native_strings(), subtype_strategy,
+        elements |= st.tuples(native_strings, subtype_strategy,
                               array_shapes(max_dims=2, max_side=2))
     return st.lists(elements=elements, min_size=min_size, max_size=max_size,
                     unique_by=lambda d: d[0])
 
 
 @st.defines_strategy
-def nested_dtypes(subtype_strategy=scalar_dtypes(),
-                  max_leaves=10, max_itemsize=None):
+def nested_dtypes(
+    subtype_strategy=scalar_dtypes(),  # type: st.SearchStrategy[np.dtype]
+    max_leaves=10,  # type: int
+    max_itemsize=None,  # type: int
+):
+    # type: (...) -> st.SearchStrategy[np.dtype]
     """Return the most-general dtype strategy.
 
     Elements drawn from this strategy may be simple (from the
@@ -498,7 +578,6 @@ def nested_dtypes(subtype_strategy=scalar_dtypes(),
     :func:`array_dtypes` with ``allow_subarrays=True``. Subdtypes in an
     array dtype may be nested to any depth, subject to the max_leaves
     argument.
-
     """
     return st.recursive(subtype_strategy,
                         lambda x: array_dtypes(x, allow_subarrays=True),
